#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#ifdef DEBUG_MEM

#include "debug.h"
#include "mem.h"



typedef struct meminfo meminfo;

struct meminfo {
  void *vpMemory;
  char *sFile;
  size_t szSize;
  unsigned uLine;
  unsigned uFreeAttempts;
  meminfo *miNext;
};

/* the head of the memory block list */
meminfo *miTop = NULL;

/**
 * mem_malloc
 *
 * Drop in replacement for malloc, handles logging
 *
 * @param szSize Size of block to allocate
 * @param sFile File in which memory is being allocated 
 * @param uLine Line on which memory is being allocated
 * @return Pointer to allocated memory or NULL on failure 
 */
void *mem_malloc(size_t szSize, const char *sFile, unsigned uLine)
{
  meminfo *miNew;

  ASSERT(szSize > 0);

  /* **** factor out **** */
  /* create a block for new entry */
  miNew = (meminfo *)malloc(sizeof(meminfo));
  if(!miNew) {
    fprintf(stderr, "mem.c: failed to allocate memory");
    exit(EXIT_FAILURE);
  }

  /* fill in block */
  miNew->uLine = uLine;
  miNew->uFreeAttempts = 0;
  /* don't use our strdup function, else we'd go into a infinite mem checking
     loop */
  miNew->sFile = (char *)malloc(strlen(sFile) +1);
  if(!miNew->sFile) {
    fprintf(stderr, "mem.c: failed to allocate memory");
    exit(EXIT_FAILURE);
  }
  strcpy(miNew->sFile, sFile);

  miNew->miNext = NULL;
  /* **** **** */

  /* possible IMPROVE */
  /* we do not have to check the return value of this, as the main code
     itself should be checking it */
  miNew->vpMemory = malloc(szSize);
  miNew->szSize = szSize;


  /* **** factor out **** */
  /* link into the list */
  if(NULL == miTop) {
    miTop = miNew;
  } else {
    meminfo *miLast = miTop;

    while(miLast->miNext != NULL) {
      miLast = miLast->miNext;
    }
    miLast->miNext = miNew;
  }
  /* **** **** */

  return miNew->vpMemory;
}

/**
 * mem_calloc
 *
 * Drop in replacement for calloc, handles logging
 *
 * @param sznMeb number of blocks to allocate
 * @param szSize Size of block to allocate
 * @return Pointer to allocated memory or NULL on failure 
 */
void *mem_calloc(size_t szNmeb, size_t szSize, const char *sFile, unsigned uLine)
{
  meminfo *miNew;

  ASSERT(szSize > 0);

  /* **** factor out **** */
  /* create a block for new entry */
  miNew = (meminfo *)malloc(sizeof(meminfo));
  if(!miNew) {
    fprintf(stderr, "mem.c: failed to allocate memory");
    exit(EXIT_FAILURE);
  }

  /* fill in block */
  miNew->uLine = uLine;
  miNew->uFreeAttempts = 0;
  /* don't use our strdup function, else we'd go into a infinite mem checking
     loop */
  miNew->sFile = (char *)malloc(strlen(sFile) +1);
  if(!miNew->sFile) {
    fprintf(stderr, "mem.c: failed to allocate memory");
    exit(EXIT_FAILURE);
  }
  strcpy(miNew->sFile, sFile);

  miNew->miNext = NULL;
  /* **** **** */

  /* possible IMPROVE */
  /* we do not have to check the return value of this, as the main code
     itself should be checking it */
  miNew->vpMemory = calloc(szNmeb, szSize);
  miNew->szSize = szSize * szNmeb;


  /* **** factor out **** */
  /* link into the list */
  if(NULL == miTop) {
    miTop = miNew;
  } else {
    meminfo *miLast = miTop;

    while(miLast->miNext != NULL) {
      miLast = miLast->miNext;
    }
    miLast->miNext = miNew;
  }
  /* **** **** */

  return miNew->vpMemory;
}

/**
 * mem_free
 *
 * Drop in replacement for free, handles un-logging
 *
 * @param vpMemPtr pointer to memory to free
 */
void mem_free(void *vpMemPtr, const char *sFile, unsigned uLine)
{
  meminfo *miLoop = miTop;

  /* cannot free something that has not be malloced */
  ASSERT(miTop != NULL);
  ASSERT(vpMemPtr);

  /* special case - first entry */
  if(miLoop->vpMemory == vpMemPtr) {    

    miLoop->vpMemory = NULL;

    miLoop->uFreeAttempts++;
    if(miLoop->uFreeAttempts > 1) {
      fprintf(stderr, "Attempt to free() something previously freed on %s %u\n",
              sFile, uLine);
    }
    return;
  }

  while(miLoop) {
    if(miLoop->miNext && (miLoop->miNext->vpMemory == vpMemPtr)) {

      miLoop->vpMemory = NULL;

      miLoop->uFreeAttempts++;
      if(miLoop->uFreeAttempts > 1) {
        fprintf(stderr, "Attempt to free() something previously freed on %s %u\n",
                sFile, uLine);
      }
      return;
    }
    miLoop = miLoop->miNext;
  }

  fprintf(stderr, "Attempt to free() something not malloc()ed on %s %u\n",
          sFile, uLine);
}

/**
 * mem_dump
 * 
 * output all unfreed memory blocks
 */
void mem_dump(void)
{
  meminfo *miLoop = miTop;
  size_t szTotal = 0;

  if(NULL == miLoop) {
    fprintf(stderr, "No unfreed memory, I'm impressed\n");
    return;
  }

  while(miLoop) {
    if(miLoop->uFreeAttempts == 0) {
      fprintf(stderr, "UNFREED: %s %u\n", miLoop->sFile, miLoop->uLine);
      szTotal += miLoop->szSize;
    }
    miLoop = miLoop->miNext;
  }
  fprintf(stderr, "Total waste: %d bytes\n", szTotal);
}

#endif /* DEBUG_MEM */
